# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

import metricsoperator.metrics.app as apps
import metricsoperator.metrics.network as network
import metricsoperator.metrics.perf as perf
import metricsoperator.metrics.storage as storage

metrics = {
    "io-sysstat": storage.io_sysstat,
    "network-osu-benchmark": network.network_osu_benchmark,
    "network-netmark": network.network_netmark,
    "perf-sysstat": perf.perf_sysstat,
    "io-fio": storage.io_fio,
    "app-lammps": apps.app_lammps,
    "app-amg": apps.app_amg,
}


def get_metric(name):
    """
    Get a named metric parser.
    """
    metric = metrics.get(name)
    if not metric:
        raise ValueError(f"Metric {name} does not have a known parser")
    return metric



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)


import re

import metricsoperator.utils as utils
from metricsoperator.metrics.base import MetricBase


class app_amg(MetricBase):
    """
    Parse AMG into nice JSON :)
    """

    container_name = "launcher"

    @property
    def pod_prefix(self):
        return f"{self.name}-l-0"

    def parse_log(self, lines):
        """
        Given lines of output, parse and return json
        """
        # Get the log metadata
        metadata = self.get_log_metadata(lines)

        # Split lines by separator
        results = []
        sections = self.get_log_sections(lines)
        for section in sections:
            if not section.strip():
                continue
            section = section.split("\n")
            # Don't lose indentation, meaningful!
            section = [x for x in section if x.strip()]
            results.append(parse_amg(section))
        return {"data": results, "metadata": metadata, "spec": self.spec}


def parse_key_value_pair(line, sep="="):
    """
    Parse a key value pair, ensuring it is slugified too.
    """
    key, value = line.split(sep, 1)
    key = utils.slugify(key.strip())
    if not key:
        return
    return key, value.strip()


def parse_params(lines):
    """
    Parse a section.

    This is a generic parsing strategy that works with the test case.
    If we need to create problem-specific parsers we can do that.
    """
    # The title here is the first section header
    params = {}
    section = None
    while lines:
        line = lines.pop(0)

        # if we find an = with a space prefix, add directly as a data attribute
        if "=" in line and line.startswith(" "):
            pair = parse_key_value_pair(line)
            if not pair:
                continue
            if section:
                params[section][pair[0]] = pair[1]
            else:
                params[pair[0]] = pair[1]

        # If we find an = and the line is on equal level, it's a standalone attribute
        # OR an indented parameter of the same section
        elif ("=" in line and not line.startswith(" ")) or (
            ":" in line and line.startswith("   ")
        ):
            pair = parse_key_value_pair(line)
            if not pair:
                continue
            print(f"Adding {pair} to section {section}")
            params[pair[0]] = pair[1]

        # If we find a colon, it's a section header
        elif ":" in line:
            parts = line.strip().split(":", 1)
            parts = [x.strip() for x in parts if x.strip()]

            # If we have one piece, it's a section name
            if len(parts) == 1:
                section = parts[0]

                # Slugify!
                section = utils.slugify(section)
                params[section] = {}
                print(f"Adding new section {section}")

            # If we have two pieces, it's a separate key/value pair
            elif len(parts) > 1:
                pair = parse_key_value_pair(line, sep=":")
                if not pair:
                    continue
                params[pair[0]] = pair[1]
                print(f"Adding {pair} to section {section}")

    return params


separator = "============================================="


def parse_amg(lines):
    """
    Parse the AMG section. This is notably different, but original credit for parsing this goes to:
    https://github.com/flux-framework/flux-k8s/blob/canopie22-artifacts/canopie22-artifacts/amg/process_amg.py
    with credit to Dan Milroy.
    """
    entry = {}

    # Prepare to split into subset
    def get_subset(lines):
        subset = []
        while lines and "======" not in lines[0]:
            subset.append(lines.pop(0))
        return subset

    while lines:
        line = lines.pop(0)

        # This is the first section, driver parameters
        if re.search("Running with these driver", line, re.IGNORECASE):
            # Get the subset up to the next section
            section = get_subset(lines)

            # Each subsection header has TITLE:
            entry["driver_params"] = parse_params(section)
            continue

        if re.search("Generate Matrix", line, re.IGNORECASE):
            lines.pop(0)  # get rid of extra header
            section = get_subset(lines)
            entry["generate_matrix"] = parse_params(section)
            continue

        if re.search("Vector Setup", line, re.IGNORECASE):
            lines.pop(0)  # get rid of extra header
            section = get_subset(lines)
            entry["vector_setup"] = parse_params(section)
            continue

        if re.search("Setup Time", line, re.IGNORECASE):
            lines.pop(0)  # get rid of extra header
            section = get_subset(lines)
            entry["problem_setup"] = parse_params(section)
            continue

        if re.search("Solve Time", line, re.IGNORECASE):
            lines.pop(0)  # get rid of extra header
            section = get_subset(lines)
            entry["solve_time"] = parse_params(section)
            continue

    return entry



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

import json
import re

from .base import MetricBase

# Known section headers
headers = [
    "CPU STATISTICS TASK",
    "CPU STATISTICS CHILD",
    "IO STATISTICS",
    "POLICY",
    "PAGEFAULTS TASK",
    "PAGEFAULTS CHILD",
    "STACK UTILIZATION",
    "THREADS TASK",
    "THREADS CHILD",
    "KERNEL TABLES",
    "TASK SWITCHING",
]


header_regex = "(%s)" % "|".join(headers)


class perf_sysstat(MetricBase):
    container_name = "perf-sysstat"

    @property
    def pod_prefix(self):
        return f"{self.name}-m-0"

    def parse_log(self, lines):
        """
        Given lines of output, parse and return json
        """
        # Get the log metadata
        metadata = self.get_log_metadata(lines)

        # Get command being monitored
        command = lines.split("PIDSTAT COMMAND START", 1)[-1]
        command = command.split("PIDSTAT COMMAND END", 1)[0].strip()

        # Split lines by section separator
        results = []
        sections = self.get_log_sections(lines)
        for section in sections:
            if not section.strip():
                continue

            # These will be parts for one timepoint
            parts = section.strip().split("\n")
            timepoint = {}
            while parts:
                part = parts.pop(0)
                title = part.replace(" ", "_").lower().strip()
                if re.search(header_regex, part) and parts:
                    jsondata = parts.pop(0)
                    try:
                        data = json.loads(jsondata)
                    except Exception:
                        print(f"Issue parsing {part}")
                        continue
                    if data:
                        timepoint[title] = data

            # Only add timepoint if we collected data
            if timepoint:
                results.append(timepoint)

        return {
            "data": results,
            "metadata": metadata,
            "command": command,
            "spec": self.spec,
        }



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)


import re

import metricsoperator.utils as utils
from metricsoperator.metrics.base import MetricBase


def get_first_int(match):
    return int(match.group().strip().split(" ")[0])


class app_lammps(MetricBase):
    """
    Parse LAMMPS into nice JSON :)
    """

    container_name = "launcher"

    @property
    def pod_prefix(self):
        return f"{self.name}-l-0"

    def parse_row(self, row):
        """
        Given a row of two values with spaces, parse.
        """
        row = row.split(" ", 1)
        return [x.strip() for x in row]

    def parse_log(self, lines):
        """
        Given lines of output, parse and return json
        """
        # Get the log metadata
        metadata = self.get_log_metadata(lines)

        # Split lines by separator
        results = []
        sections = self.get_log_sections(lines)
        for section in sections:
            if not section.strip():
                continue
            section = section.split("\n")
            section = [x.strip() for x in section if x.strip()]
            results.append(parse_lammps(section))
        return {"data": results, "metadata": metadata, "spec": self.spec}


def parse_lammps(lines):
    """
    Parse a raw lammps log and return data
    """
    if not isinstance(lines, list):
        lines = lines.split("\n")

    entry = {}
    while lines:
        line = lines.pop(0)

        # Figure out ranks and threads (in same line)
        match = re.search("[0-9]+ MPI tasks", line)
        if match:
            entry["ranks"] = get_first_int(match)
            continue

        match = re.search("[0-9]+ OpenMP threads", line)
        if match:
            entry["threads"] = get_first_int(match)

        match = re.search("(?P<percentage>[0-9]+[.][0-9]+)[%] CPU use with", line)
        if match:
            entry["percentage_cpu"] = float(match.groupdict()["percentage"])
            continue

        if re.search("^(NLocal|Nghost|Neighs)", line):
            category = line.split(":")[0].lower()
            line = line.split(":", 1)[-1].strip()
            ave, line = line.split("ave", 1)
            maxval, line = line.split("max", 1)
            minval, line = line.split("min", 1)
            entry.update(
                {
                    f"{category}_avg": float(ave),
                    f"{category}_min": float(minval),
                    f"{category}_max": float(maxval),
                }
            )
            hist = lines.pop(0).split(":")[-1].strip().split(" ")
            entry[f"{category}_hist"] = [int(x) for x in hist]
            continue

        if line.startswith("Total # of neighbors"):
            entry["neighbors"] = float(line.split("=")[-1].strip())
            continue

        if line.startswith("Ave neighs/atom"):
            entry["average_neighbors_per_atom"] = float(line.split("=")[-1].strip())
            continue

        if line.startswith("Neighbor list builds"):
            entry["neighbor_list_builds"] = int(line.split("=")[-1].strip())
            continue

        # This is total wall time as reported by lammps
        if line.startswith("Total wall time"):
            rawtime = line.split(":", 1)[-1].strip()
            entry["total_wall_time_raw"] = rawtime
            entry["total_wall_time_seconds"] = utils.timestr2seconds(rawtime)
            continue

        # Dimension of molecular matrix maybe?
        match = re.search(
            "(?P<x>[0-9]+) by (?P<y>[0-9]+) by (?P<z>[0-9]+) MPI processor grid",
            line,
        )

        if match:
            if "processor_grids" not in entry:
                entry["processor_grids"] = []
            new_grid = [{k: int(v)} for k, v in match.groupdict().items()]
            entry["processor_grids"].append(new_grid)
            continue

        # Number of atoms / velocities
        match = re.search("[0-9]+ atoms", line)
        if match and "atoms" not in entry:
            entry["atoms"] = get_first_int(match)
            continue
        match = re.search("[0-9]+ velocities", line)
        if match:
            entry["velocities"] = get_first_int(match)
            continue

        # reading data from CPU
        match = re.search("read_data CPU = (?P<cpu>[0-9]+[.][0-9]+) seconds", line)
        if match:
            entry["read_data_cpu_seconds"] = float(match.groupdict()["cpu"])
            continue

        match = re.search(
            "bounding box extra memory = (?P<mem>[0-9]+[.][0-9]+) MB", line
        )
        if match:
            entry["bounding_box_extra_memory_mb"] = float(match.groupdict()["mem"])
            continue

        match = re.search("replicate CPU = (?P<seconds>[0-9]+[.][0-9]+) seconds", line)
        if match:
            entry["replicate_cpu_seconds"] = float(match.groupdict()["seconds"])
            continue

        if line.startswith("Unit style"):
            entry["unit_style"] = line.split(":")[-1].strip()
            continue

        if line.startswith("Time step"):
            entry["time_step"] = float(line.split(":")[-1].strip())
            continue

        match = re.search(
            "Per MPI rank memory allocation (min/avg/max) = (?P<min_mpi_rank_memory_allocation_mb>[0-9]+[.][0-9]+) [|] (?P<avg_mpi_rank_memory_allocation_mb>[0-9]+[.][0-9]+) [|] (?P<max_mpi_rank_memory_allocation_mb>[0-9]+[.][0-9]+) Mbytes",
            line,
        )
        if match:
            [entry.update({k: int(v)}) for k, v in match.groupdict().items()]
            continue

        # If we find the embedded table with steps
        if line.startswith("Step"):
            header = [x.strip() for x in line.split(" ") if x.strip()]
            matrix = []
            line = lines.pop(0)
            while not line.startswith("Loop"):
                matrix.append([float(x.strip()) for x in line.split(" ") if x.strip()])
                line = lines.pop(0)
            entry["steps"] = {"matrix": matrix, "columns": header}

            # Here we have the loop line
            match = re.search(
                "Loop time of (?P<loop_time>[0-9]+[.][0-9]+) on (?P<loop_procs>[0-9]+) procs for (?P<loop_steps>[0-9]+) steps with (?P<loop_atoms>[0-9]+) atoms",
                line,
            )
            if match:
                values = match.groupdict()
                entry.update(
                    {
                        "loop_time": float(values["loop_time"]),
                        "loop_procs": int(values["loop_procs"]),
                        "loop_atoms": int(values["loop_atoms"]),
                    }
                )
                continue

        match = re.search(
            "Performance: (?P<performance_ns_per_day>[0-9]+[.][0-9]+) ns/day, (?P<performance_hours_per_ns>[0-9]+[.][0-9]+) hours/ns, (?P<timesteps_per_second>[0-9]+[.][0-9]+) timesteps/s",
            line,
        )
        if match:
            [entry.update({k: float(v)}) for k, v in match.groupdict().items()]
            continue

        # If we find the embedded table with times
        if line.startswith("Section"):
            header = [x.strip() for x in line.split("|") if x.strip()]

            # Line with full ----------------------
            lines.pop(0)
            matrix = []
            columns = header[1:]
            line = lines.pop(0)
            while line and line.strip():
                parts = [x.strip() for x in line.split("|")]
                _, rest = parts[0], parts[1:]
                if rest:
                    matrix.append(rest)
                if not lines:
                    break
                line = lines.pop(0)

            entry["times"] = {"matrix": matrix, "header": columns}

    return entry



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

import json
import os
import re

import yaml


def read_file(filename):
    """
    Read file into text blob
    """
    with open(filename, "r") as fd:
        text = fd.read()
    return text


def read_yaml(filename):
    """
    Read yaml into dict
    """
    with open(filename, "r") as file:
        configuration = yaml.safe_load(file)
    return configuration


def slugify(section):
    """
    Slugify into lowercase and underscores
    """
    return section.lower().replace(" ", "_").replace("-", "_")


def write_file(content, filename):
    """
    Write dict to json
    """
    with open(filename, "w") as fd:
        fd.write(content)


def write_json(obj, filename):
    """
    Write dict to json
    """
    with open(filename, "w") as fd:
        fd.write(json.dumps(obj, indent=4))


def read_json(filename):
    """
    Read json into dict
    """
    with open(filename, "r") as fd:
        content = json.loads(fd.read())
    return content


def recursive_find(base, pattern="^(laamps[.]out|lammps.*[.]out|log[.]out)$"):
    """
    Recursively find lammps output files.
    """
    for root, _, filenames in os.walk(base):
        for filename in filenames:
            if re.search(pattern, filename):
                yield os.path.join(root, filename)


def read_lines(filename):
    """
    Read lines of a file into a list.
    """
    with open(filename, "r") as fd:
        lines = fd.readlines()
    return lines


def timestr2seconds(timestr):
    """
    Given a timestring in two formats, return seconds (float).
    """
    # Minutes and seconds, MM:SS.mm
    if timestr.count(":") == 1:
        minutes, seconds = timestr.split(":")
        return (int(minutes) * 60) + float(seconds)

    # hours, minutes, seconds HH:MM:SS.mm
    elif timestr.count(":") == 2:
        hours, minutes, seconds = timestr.split(":")
        return (int(hours) * 360) + (int(minutes) * 60) + float(seconds)
    raise ValueError(f"Unrecognized time format {timestr}")



# Copyright 2023 Lawrence Livermore National Security, LLC
# (c.f. AUTHORS, NOTICE.LLNS, COPYING)

# SPDX-License-Identifier: MIT

import os

from kubernetes import client, config

import metricsoperator.metrics as mutils
import metricsoperator.utils as utils


class MetricsOperator:
    def __init__(self, yaml_file):
        """
        Given a YAML file with one or more metrics, apply
        to create it and stream logs for each metric of interest.
        """
        self._core_v1 = None
        self.yaml_file = os.path.abspath(yaml_file)
        self.spec = utils.read_yaml(self.yaml_file)
        config.load_kube_config()

    def watch(self):
        """
        Wait for (and yield parsed) metric logs.
        """
        for metric in self.spec["spec"]["metrics"]:
            parser = mutils.get_metric(metric["name"])(self.spec)
            print("Watching %s" % metric["name"])
            for pod, container in parser.logging_containers():
                yield parser.parse(pod=pod, container=container)

    def create(self):
        """
        Create the associated YAML file.
        """
        api = client.CustomObjectsApi()
        return api.create_namespaced_custom_object(
            group=self.group,
            version=self.version,
            namespace=self.namespace,
            plural=self.plural,
            body=self.spec,
        )

    @property
    def group(self):
        return self.spec["apiVersion"].split("/", 2)[0]

    @property
    def version(self):
        return self.spec["apiVersion"].split("/", 2)[1]

    @property
    def plural(self):
        return self.spec["kind"].lower() + "s"

    @property
    def namespace(self):
        return self.spec["metadata"].get("namespace") or "default"

    @property
    def name(self):
        return self.spec["metadata"]["name"]

    def delete(self):
        """
        Delete the associated YAML file.
        """
        api = client.CustomObjectsApi()
        result = api.delete_namespaced_custom_object(
            group=self.group,
            version=self.version,
            namespace=self.namespace,
            plural=self.plural,
            name=self.name,
        )
        self.wait_for_delete()
        return result

    def wait_for_delete(self):
        """
        Wait for pods to be gone (deleted)
        """
        for metric in self.spec["spec"]["metrics"]:
            parser = mutils.get_metric(metric["name"])(self.spec)
            print("Watching %s for deletion" % metric["name"])
            parser.wait_for_delete()
